@@ -3,10 +3,12 @@
 // Copyright (c) 2001-2005 David Ward
 
 #include "../Common/Common.h"
+
 #include <iostream>
 #include "../Common/Random.h"
 #include "DasherModel.h"
 #include "DasherView.h"
+#include "Parameters.h"
 
 using namespace Dasher;
 using namespace std;
@@ -84,7 +86,7 @@ m_Rootmax_max(0), m_dAddProb(0.0), m_dMaxRate(0.0) {
   LearnContext = m_pLanguageModel->CreateEmptyContext();
 
   // various settings
-  int iShift = 12;
+  // int iShift = 12;
   // m_DasherY = 1 << iShift;
   //  m_DasherOY = m_DasherY / 2;
   //  m_DasherOX = m_DasherY / 2;
@@ -141,6 +143,9 @@ void CDasherModel::HandleEvent(Dasher::CEvent *pEvent) {
       m_dMaxRate = GetLongParameter(LP_MAX_BITRATE) * GetLongParameter(LP_BOOSTFACTOR) / 100 / static_cast<double>(GetLongParameter(LP_SPEED_DIVISOR));
       m_fr.SetMaxBitrate(m_dMaxRate);
       break;
+    case BP_CONTROL_MODE: // Rebuild the model if control mode is switched on/off
+      RebuildAroundNode(Get_node_under_crosshair());
+      break;
     default:
       break;
     }
@@ -151,7 +156,6 @@ void CDasherModel::HandleEvent(Dasher::CEvent *pEvent) {
 void CDasherModel::Make_root(CDasherNode *whichchild)
 // find a new root node 
 {
-
   symbol t = m_Root->Symbol();
   if(t < m_pDasherInterface->GetAlphabet()->GetNumberTextSymbols()) {
     // Only learn if we have adaptive behaviour enabled
@@ -179,6 +183,42 @@ void CDasherModel::Make_root(CDasherNode *whichchild)
   m_Rootmin = m_Rootmin + (range * m_Root->Lbnd()) / (int)GetLongParameter(LP_NORMALIZATION);
 }
 
+void CDasherModel::ClearRootQueue() {
+  while(oldroots.size() > 0) {
+    if(oldroots.size() > 1) {
+      oldroots[0]->OrphanChild(oldroots[1]);
+    }
+    else {
+      oldroots[0]->OrphanChild(m_Root);
+    }
+    delete oldroots[0];
+    oldroots.pop_front();
+  }
+}
+
+void CDasherModel::RecursiveMakeRoot(CDasherNode *pNewRoot) {
+  if(!pNewRoot)
+    return;
+
+  if(pNewRoot == m_Root)
+    return;
+
+  // FIXME - we really ought to check that pNewRoot is actually a
+  // descendent of the root, although that should be guaranteed
+
+  if(!pNewRoot->NodeIsParent(m_Root))
+    RecursiveMakeRoot(pNewRoot->Parent());
+
+  Make_root(pNewRoot);
+}
+
+void CDasherModel::RebuildAroundNode(CDasherNode *pNode) {
+  RecursiveMakeRoot(pNode);
+  ClearRootQueue();
+  pNode->Delete_children();
+  pNode->m_pNodeManager->PopulateChildren(pNode);
+}
+
 void CDasherModel::Reparent_root(int lower, int upper) {
 
   /* Change the root node to the parent of the existing node
@@ -310,7 +350,7 @@ void CDasherModel::SetContext(std::string &sNewContext) {
 
   double dFraction( 1 - (1 - m_Root->MostProbableChild() / static_cast<double>(GetLongParameter(LP_NORMALIZATION))) / 2.0 );
 
-  int iWidth( GetLongParameter(LP_MAX_Y) / (2.0*dFraction) );
+  int iWidth( static_cast<int>( (GetLongParameter(LP_MAX_Y) / (2.0*dFraction)) ) );
 
   m_Rootmin = GetLongParameter(LP_MAX_Y) / 2 - iWidth / 2;
   m_Rootmax = GetLongParameter(LP_MAX_Y) / 2 + iWidth / 2;
@@ -693,7 +733,8 @@ double CDasherModel::Plan_new_goto_coords(int iRxnew, myint mousey, int *iSteps,
     double s = (log(2.0) * 2 / log( (STEPDENOM*1.0)/(m_Stepnum*1.0)) ) / 4096;
 
     double alpha = 2 * (2 * s);
-    int alternateSteps = int(alpha * distance);
+    int alternateSteps = int(alpha * abs(distance));
+
     // Take log of iRxnew to base ( STEPDENOM / STEPNUM ):
     if ( STEPDENOM > m_Stepnum && m_Stepnum > 0 ) { // check that the following loop will terminate.
       //cout << "iRxnew is " << iRxnew << " and ZOOMDENOM is" << ZOOMDENOM << endl;
@@ -709,6 +750,7 @@ double CDasherModel::Plan_new_goto_coords(int iRxnew, myint mousey, int *iSteps,
         }
       }
     }
+
     // Done taking log of iRxnew. 
     if (alternateSteps > *iSteps) {
       *iSteps = alternateSteps;
